#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <termios.h>
#include "ansi.h"
#include "crc32.h"
#include "lgc.h"
#include "output.h"
#include "game.h"

#define PKT_CONTROL		0x80
#define CTRL_CONNECT		0
#define	CTRL_DISCONNECT		1
#define CTRL_SETUP_REQUEST	2
#define CTRL_SETUP_NOTIFY	3
#define CTRL_RESET_REQUEST	4
#define	CTRL_RESET_NOTIFY	5
#define CTRL_SYNC_REQUEST	6
#define CTRL_SYNC_NOTIFY	7
#define STOPCODE_LEN		16

typedef struct NetNode {
  int                      clientNo;
  int                      fd;
} NetNode;

char                     gameSetup[256];
int                      gameClient,
                         gameInProgress,
                         initialNodes,
                         nodeAddr,
                         maxAddr;
NetNode                  netNode[MAX_NETNODES];

void
networkStart()
{
  int                      i,
                           j,
                           newPacket = 1,
                           CRCbyte = 0,
                           broadcast = 0,
                           esc = 0,
                           matchLen = 0,
                           iBufI,
                           iBufN;
  byte                     c,
                           pktDst = 0,
                           iBuf[COMM_BUFSIZE + 64],
                           ctlMsg[64],
                           ctlMsgLen;
  byte                     stopCode[STOPCODE_LEN];
  fd_set                   allSet,
                           rSet;

  for (i = 0; i < MAX_NETNODES; i++)
    netNode[i].clientNo = netNode[i].fd = -1;
  maxAddr = -1;
  if (optDOS) {
    for (i = 0; i < STOPCODE_LEN; i++)
      if (i % 2)
	stopCode[i] = 0xff;
      else
	stopCode[i] = 0;
    flock(STDOUT_FILENO, LOCK_EX);
    for (i = 0; i < 3; i++) {
      DOSwriteByte(PKT_START);
      DOSwriteByte(gameClient);
      char                     buf[256];

      sprintf(buf, "%d %d %d %s", gameInProgress, initialNodes, nodeAddr, gameSetup);
      DOSwriteString(buf, strlen(buf) + 1);
      DOSendPacket();
    }
    flock(STDOUT_FILENO, LOCK_UN);
  } else {
    for (i = 0; i < STOPCODE_LEN; i++)
      stopCode[i] = 24;
    _werase(msgW);
    wrefresh(msgW);
    msgStr("<C %d>**</C> Virtual network activated\n", FG_STAR);
    raw();
  }
  FD_ZERO(&allSet);
  FD_SET(STDIN_FILENO, &allSet);
  FD_SET(fd, &allSet);
  for (;;) {
    rSet = allSet;
    select(fd + 1, &rSet, NULL, NULL, NULL);
    if (FD_ISSET(fd, &rSet))
      while (!clientRead());
    while (nMsgsWaiting) {
      CRC32                    crc;

      getMessage(&msg);
      switch (msg.type) {
      case MSG_PING:
	i = sizeof(int);
	j = CMD_PING;
	write(fd, &i, sizeof(int));
	write(fd, &j, sizeof(int));
	break;
      case MSG_CONNECT:
	i = *(int *) msg.data;
	j = *((int *) msg.data + 1);
	refreshClients(0);
	netNode[i].clientNo = j;
	netNode[i].fd = open(client[j].line, O_WRONLY);
	maxAddr = max(maxAddr, i);
	crc.addByte(PKT_CONTROL | nodeAddr);
	crc.addByte(CTRL_CONNECT);
	crc.addByte(i);
	ctlMsgLen = 0;
	ctlMsg[ctlMsgLen++] = PKT_CONTROL | nodeAddr;
	ctlMsg[ctlMsgLen++] = crc.CRC() & 0xff;
	ctlMsg[ctlMsgLen++] = CTRL_CONNECT;
	ctlMsg[ctlMsgLen++] = i;
	ctlMsg[ctlMsgLen++] = FRAME_CHAR;
	ctlMsg[ctlMsgLen++] = (crc.CRC() & 0x7f00) >> 8;
	flock(STDOUT_FILENO, LOCK_EX);
	for (i = 0; i < 3; i++)
	  write(STDOUT_FILENO, ctlMsg, ctlMsgLen);
	flock(STDOUT_FILENO, LOCK_UN);
	break;
      case MSG_DISCONNECT:
	i = *(int *) msg.data;
	close(netNode[i].fd);
	netNode[i].clientNo = netNode[i].fd = -1;
	if (i == maxAddr)
	  for (maxAddr--; maxAddr >= 0; maxAddr--)
	    if (netNode[i].clientNo >= 0)
	      break;
	crc.addByte(PKT_CONTROL | nodeAddr);
	crc.addByte(CTRL_DISCONNECT);
	crc.addByte(i);
	ctlMsgLen = 0;
	ctlMsg[ctlMsgLen++] = PKT_CONTROL | nodeAddr;
	ctlMsg[ctlMsgLen++] = crc.CRC() & 0xff;
	ctlMsg[ctlMsgLen++] = CTRL_DISCONNECT;
	ctlMsg[ctlMsgLen++] = i;
	ctlMsg[ctlMsgLen++] = FRAME_CHAR;
	ctlMsg[ctlMsgLen++] = (crc.CRC() & 0x7f00) >> 8;
	flock(STDOUT_FILENO, LOCK_EX);
	for (i = 0; i < 3; i++)
	  write(STDOUT_FILENO, ctlMsg, ctlMsgLen);
	flock(STDOUT_FILENO, LOCK_UN);
	break;
      }
    }
    if (FD_ISSET(STDIN_FILENO, &rSet)) {
      iBufI = 0;
      iBufN = read(STDIN_FILENO, iBuf, COMM_BUFSIZE);
      while (iBufI < iBufN) {
	c = iBuf[iBufI++];
	if (c == stopCode[matchLen]) {
	  if (++matchLen == STOPCODE_LEN)
	    return;
	} else
	  matchLen = 0;
	// packet header?
	if (newPacket) {
	  pktDst = c & 0x3f;
	  broadcast = pktDst == nodeAddr;
	  c = (c & ~0x3f) | nodeAddr;
	  if (broadcast) {
	    for (i = 0; i <= maxAddr; i++)
	      if (netNode[i].fd >= 0)
		flock(netNode[i].fd, LOCK_EX);
	  } else {
	    if (pktDst <= maxAddr && netNode[pktDst].fd >= 0)
	      flock(netNode[pktDst].fd, LOCK_EX);
	  }
	}
	// write the byte
	if (broadcast) {
	  for (i = 0; i <= maxAddr; i++)
	    if (netNode[i].fd >= 0)
	      write(netNode[i].fd, &c, 1);
	} else {
	  if (pktDst <= maxAddr && netNode[pktDst].fd >= 0)
	    write(netNode[pktDst].fd, &c, 1);
	}
	if (newPacket) {
	  newPacket = 0;
	  CRCbyte = 1;
	  continue;
	}
	if (CRCbyte) {
	  CRCbyte = 0;
	  continue;
	}
	// end of packet?
	if (esc) {
	  esc = 0;
	  if (c != FRAME_CHAR) {
	    if (broadcast) {
	      for (i = 0; i <= maxAddr; i++)
		if (netNode[i].fd >= 0)
		  flock(netNode[i].fd, LOCK_UN);
	    } else {
	      if (pktDst <= maxAddr && netNode[pktDst].fd >= 0)
		flock(netNode[pktDst].fd, LOCK_UN);
	    }
	    newPacket = 1;
	  }
	} else if (c == FRAME_CHAR)
	  esc = 1;
      }
    }
  }
}

void
networkEnd()
{
  int                      i,
                           cmd = CMD_DISCONNECT;

  for (i = 0; i <= maxAddr; i++)
    if (netNode[i].fd >= 0)
      close(netNode[i].fd);
  if (!optDOS) {
    tcsetattr(STDIN_FILENO, TCSAFLUSH, &saveTermios);
    ncursesInit();
  }
  refreshClients();
  msgStr("<C %d>**</C> <C %d>DISCONNECTED</C> from <C %d>%s</C>'s network\n", FG_STAR, FG_CAPS, clientColor(client[client[clientNo].ccid].name), client[client[clientNo].ccid].name);
  request(&cmd, sizeof(int));
  refreshClients();
  writeClients(curChannel);
}
